import math
from pyglet.gl import glClearColor
from ververser import GameWindow


class Game:

    def __init__( self, game_window : GameWindow ):
        self.game_window = game_window
        self.total_time = 0

    def update( self, dt ):
        self.total_time += dt
        g = 0.25 + 0.25 * (math.sin( self.total_time ) + 1)
        glClearColor( 0, g, 0, 1.0 )

    def draw( self ):
        ...



from typing import Optional
from ververser import GameWindow, import_script

game = import_script( 'game.py' )
Game = game.Game


# --------
# This is just some boilerplate that makes it easier to use a custom Game class

_GAME : Optional[ Game ] = None

def vvs_init( game_window : GameWindow ):
    global _GAME
    _GAME = Game( game_window )

def vvs_update( dt ):
    _GAME.update( dt )

def vvs_draw():
    _GAME.draw()


import math
from pyglet.gl import glClearColor
from typing import Optional
from ververser import GameWindow


class Game:

    def __init__( self, game_window : GameWindow ):
        self.game_window = game_window
        self.total_time = 0

    def update( self, dt ):
        self.total_time += dt
        g = 0.25 + 0.25 * ( math.sin( self.total_time ) + 1 )
        glClearColor( 0, g, 0, 1.0 )

    def draw( self ):
        ...


# --------
# This is just some boilerplate that makes it easier to use a custom Game class

_GAME : Optional[ Game ] = None

def vvs_init( game_window : GameWindow ):
    global _GAME
    _GAME = Game( game_window )

def vvs_update( dt ):
    _GAME.update( dt )

def vvs_draw():
    _GAME.draw()


import logging
from pathlib import Path
from runpy import run_path


logger = logging.getLogger(__name__)


class MainScript:

    def __init__( self, file_path : Path, game_window ):
        self.file_path = file_path
        self.data_module = run_path( str( file_path ) )

        self.vvs_init( game_window )

    def vvs_init( self, game_window ):
        f_init = self.data_module.get( 'vvs_init' )
        if f_init :
            f_init( game_window )

    def vvs_update( self, dt ):
        f_update = self.data_module.get( 'vvs_update' )
        if f_update :
            f_update( dt )

    def vvs_draw( self ):
        f_draw = self.data_module.get( 'vvs_draw' )
        if f_draw:
            f_draw()


def load_main_script( script_path : Path, game_window ) -> MainScript:
    main_script = None
    try :
        main_script = MainScript( script_path, game_window )
    except Exception as e :
        logger.exception( f'Encountered an error during loading of main script "{script_path}". Exception: {e}' )
    return main_script



from enum import Enum, auto
import logging
from typing import Any

from ververser.file_watcher import FileWatcher


logger = logging.getLogger(__name__)


class LoadStatus( Enum ):
    NOT_CHANGED = auto()
    RELOADED = auto()
    FAILED = auto()


class ReloadingAsset:

    def __init__(self, f_load_asset, file_path):
        self.f_load_asset = f_load_asset
        self.file_watcher = FileWatcher( file_path, record_now = False )
        self.asset = None
        self.try_reload()

    def __getattr__( self, name : str ) -> Any:
        return getattr( self.asset, name )

    def __bool__(self) -> bool:
        return bool( self.asset )

    def try_reload( self ) -> None:
        if not self.file_watcher.is_file_updated():
            self.reload_status = LoadStatus.NOT_CHANGED
            return
        asset_path = self.file_watcher.file_path
        try:
            self.asset = self.f_load_asset( asset_path )
        except BaseException as e :
            logger.exception( f'Encountered an error during loading of asset from file "{asset_path}". Exception: {e}' )
            self.reload_status = LoadStatus.FAILED
            self.asset = None
            return
        self.reload_status =  LoadStatus.RELOADED
        return

    def get( self ) -> Any:
        return self.asset





from ververser.global_game_window import get_global_game_window
from ververser.script import Script


def import_script( script_path : str ) -> Script:
    return get_global_game_window().asset_manager.load( script_path )



from typing import Optional, Any

from ververser.game_window import GameWindow


_GAME_WINDOW = Optional[ Any ]


def get_global_game_window() -> GameWindow:
    assert _GAME_WINDOW, 'Global game window not set'
    return _GAME_WINDOW


def set_global_game_window( game_window : GameWindow ) -> None:
    global _GAME_WINDOW
    _GAME_WINDOW = game_window



from pathlib import Path
from runpy import run_path


class Script:

    def __init__( self, file_path : Path ):
        self.file_path = file_path
        self.data_module = run_path( str( file_path ) )

    def __getattr__( self, name ):
        return self.data_module.get( name )


def load_script( script_path : Path ) -> Script:
    return Script( script_path )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ververser',
 'ververser.examples.1_minimal_setup',
 'ververser.examples.1_minimal_setup.content',
 'ververser.examples.2_game_class',
 'ververser.examples.2_game_class.content',
 'ververser.examples.3_reloading_script_imports',
 'ververser.examples.3_reloading_script_imports.content']

package_data = \
{'': ['*']}

install_requires = \
['pyglet>=2.0.5,<3.0.0']

setup_kwargs = {
    'name': 'ververser',
    'version': '0.1.13',
    'description': 'A lightweight wrapper around pyglet that allows hot-reloading of content.',
    'long_description': 'None',
    'author': 'Berry van Someren',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Any, Callable

from ververser.reloading_asset import ReloadingAsset, LoadStatus
from ververser.main_script import load_main_script, MainScript
from ververser.multi_file_watcher import MultiFileWatcher
from ververser.script import load_script, Script


ScriptType = MainScript | Script
ContentType = ScriptType | ReloadingAsset

AssetLoaderType = Callable[ [ Path ], Any ]

EXPECTED_SCRIPT_EXTENSION = '.py'
EXPECTED_MAIN_SCRIPT_NAME = Path( 'main.py' )


def is_script_path( path : Path ) -> bool:
    return path.suffix == EXPECTED_SCRIPT_EXTENSION


class ContentManager:

    """
    The asset manager works mainly on two categories of content:
    1) scripts
        When a single script is updated we will not try to do anything clever.
        We basically reinitialize the entire game to make sure all scripts are reloaded.
        In future versions of ververser we can always make this more intelligent.
        The game window is responsible for checking if it needs to reinitialize.
    2) non-scripts
        Non scripts can be anything, i.e. Meshes, Shader or Textures
        Reloading these, does not require reloading the entire game,
        We will just replace them while we keep running.

    Because scripts are dealt with separately,
    we will make a hard distinction between these two categories,
    and refer to non-script content as "assets".
    """

    def __init__( self, content_folder_path : Path ):
        self.content_folder_path = content_folder_path
        self.script_watcher = MultiFileWatcher( 'Script Watcher' )

        self.asset_loaders : list[ tuple[ str, AssetLoaderType ] ] = []
        self.reloading_assets : list[ ReloadingAsset ] = []

        self.register_asset_loader( '.py', load_script )
        self.script_watcher.add_file_watch( self.make_content_path_complete( EXPECTED_MAIN_SCRIPT_NAME ) )

    # ======== Generic functions ========

    def make_content_path_complete( self, asset_path : Path ) -> Path:
        return self.content_folder_path / asset_path

    def exists( self, asset_path ) -> bool:
        complete_asset_path = self.make_content_path_complete( asset_path )
        return complete_asset_path.is_file()

    def load( self, content_path ) -> ContentType:
        absolute_content_path = self.make_content_path_complete( content_path.strip() )
        if not self.exists( absolute_content_path ):
            raise ValueError( f'Content file does not exist. File path: "{absolute_content_path}"' )
        if is_script_path( absolute_content_path ):
            return self._load_script( absolute_content_path )
        else:
            return self._load_asset( absolute_content_path )

    # ======== Script related functions ========

    def is_any_script_updated( self ) -> bool:
        return self.script_watcher.is_any_file_updated()

    def load_main_script( self, game_window ) -> MainScript:
        absolute_script_path = self.make_content_path_complete( EXPECTED_MAIN_SCRIPT_NAME )
        assert self.exists( absolute_script_path ), f'Could not load asset. File path: "{EXPECTED_MAIN_SCRIPT_NAME}"'
        main_script = load_main_script( absolute_script_path, game_window )
        self.script_watcher.add_file_watch( absolute_script_path )
        return main_script

    def _load_script( self, absolute_script_path : Path ) -> Script:
        self.script_watcher.add_file_watch( absolute_script_path )
        return load_script( absolute_script_path )

    # ======== Asset related functions ========

    def _load_asset( self, absolute_asset_path ) -> ReloadingAsset :
        asset_loader = self.get_asset_loader_for_file( absolute_asset_path )
        reloading_asset = ReloadingAsset(
            f_load_asset = asset_loader,
            file_path = absolute_asset_path
        )
        self.reloading_assets.append( reloading_asset )
        reloading_asset.try_reload()
        return reloading_asset

    def try_reload_assets( self ) -> LoadStatus:
        overall_load_status = LoadStatus.NOT_CHANGED
        for reloading_asset in self.reloading_assets:
            reloading_asset.try_reload()
            load_status = reloading_asset.reload_status
            if load_status == LoadStatus.RELOADED:
                overall_load_status = LoadStatus.RELOADED
            if load_status == LoadStatus.FAILED:
                return LoadStatus.FAILED
        return overall_load_status

    def register_asset_loader( self, postfix : str, f_load_asset : AssetLoaderType ) -> None:
        self.asset_loaders.append( ( postfix, f_load_asset ) )

    def get_asset_loader_for_file( self, file_path : Path ) -> AssetLoaderType:
        # reverse search through all registered loaders
        # this way. newest registered loaders overrule older ones
        for postfix, asset_loader in reversed( self.asset_loaders ):
            if str( file_path ).endswith( postfix ):
                return asset_loader
        raise KeyError( f'No asset loader found for file_path: "{file_path}". Known loaders: {self.asset_loaders}' )




